# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constant Value Network that always predicts a constant value."""

import gin
import tensorflow as tf

from tf_agents.networks import network
from tf_agents.utils import nest_utils


@gin.configurable
class ConstantValueNetwork(network.Network):
  """Constant value network that predicts zero per batch item."""

  def __init__(self, input_tensor_spec, constant_output_val=0, name=None):
    """Creates an instance of `ConstantValueNetwork`.

    Network supports calls with shape outer_rank + observation_spec.shape. Note
    outer_rank must be at least 1.

    Args:
      input_tensor_spec: A `tensor_spec.TensorSpec` or a tuple of specs
        representing the input observations.
      constant_output_val: A constant scalar value the network will output.
      name: A string representing name of the network.

    Raises:
      ValueError: If input_tensor_spec is not an instance of network.InputSpec.
    """
    super().__init__(
        input_tensor_spec=input_tensor_spec, state_spec=(), name=name)

    self._constant_output_val = constant_output_val

  def call(self, inputs, step_type=None, network_state=(), training=False):
    _ = (step_type, training)
    shape = nest_utils.get_outer_shape(inputs, self._input_tensor_spec)
    constant = tf.constant(self._constant_output_val, tf.float32)
    return tf.fill(shape, constant), network_state



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for storing and processing best trajectories."""

import dataclasses
import json
from typing import Dict, List

import tensorflow as tf

from compiler_opt.rl import constant


@dataclasses.dataclass(frozen=True)
class BestTrajectory:
  reward: float
  action_list: List[int]


class BestTrajectoryRepo:
  """Class for storing and processing best trajectory related operations."""

  def __init__(self, action_name: str):
    """Constructor.

    Args:
      action_name: action name of the trajectory, used for extracting action
        list from tensorflow.SequenceExample.
    """
    # {module_name: {identifier: best trajectory}}
    self._best_trajectories: Dict[str, Dict[str, BestTrajectory]] = {}
    self._action_name: str = action_name

  @property
  def best_trajectories(self) -> Dict[str, Dict[str, BestTrajectory]]:
    return self._best_trajectories.copy()

  def sink_to_json_file(self, path: str):
    with tf.io.gfile.GFile(path, 'w') as f:
      json.dump(self._best_trajectories, f, cls=constant.DataClassJSONEncoder)

  def load_from_json_file(self, path: str):
    with tf.io.gfile.GFile(path, 'r') as f:
      data = json.load(f)
    for k, v in data.items():
      if v:
        self._best_trajectories[k] = {
            sub_k: BestTrajectory(**sub_v) for sub_k, sub_v in v.items()
        }

  def sink_to_csv_file(self, path: str):
    """sink to csv file format consumable by compiler."""
    with tf.io.gfile.GFile(path, 'w') as f:
      for k, v in self._best_trajectories.items():
        for sub_k, sub_v in v.items():
          f.write(','.join([k, sub_k] + [str(x) for x in sub_v.action_list]) +
                  '\n')

  def combine_with_other_repo(self, other: 'BestTrajectoryRepo'):
    """combine and update with other best trajectory repo."""
    for k, v in other.best_trajectories.items():
      if k not in self._best_trajectories:
        self._best_trajectories[k] = v
        continue
      for sub_k, sub_v in v.items():
        if sub_v.reward < self._best_trajectories[k][sub_k].reward:
          self._best_trajectories[k][sub_k] = sub_v

  def update_if_better_trajectory(self, module_name: str, identifier: str,
                                  reward: float, trajectory: bytes):
    """update with incoming trajectory if the reward is lower.

    Args:
      module_name: module name of the trajectory.
      identifier: identifier of the trajectory within module.
      reward: reward of the trajectory.
      trajectory: trajectory in the format of serialized SequenceExample.
    """
    if module_name not in self._best_trajectories:
      self._best_trajectories[module_name] = {}
    if (identifier not in self._best_trajectories[module_name] or
        self._best_trajectories[module_name][identifier].reward > reward):
      example = tf.train.SequenceExample.FromString(trajectory)
      action_list = [
          x.int64_list.value[0]
          for x in example.feature_lists.feature_list[self._action_name].feature
      ]
      self._best_trajectories[module_name][identifier] = BestTrajectory(
          reward=reward, action_list=action_list)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate sparse buckets - aka "vocab".

Generate numerical features' X (1000) quantile based on their distributions.
"""

import math
import multiprocessing as mp
import os
from typing import Callable, Dict, List, Iterable

from absl import app
from absl import flags
from absl import logging
import gin

import numpy as np
import tensorflow as tf

from compiler_opt.rl import registry

flags.DEFINE_string('input', None,
                    'Path to input file containing tf record datasets.')
flags.DEFINE_string('output_dir', None,
                    'Path to output directory to store quantiles per feature.')
flags.DEFINE_float('sampling_fraction', 1.0,
                   'Fraction to downsample input data.', 0.0, 1.0)
flags.DEFINE_integer(
    'parallelism', None, 'Number of parallel processes to spawn.'
    'Each process does vocab generation for each feature.', 1)
flags.DEFINE_integer('num_buckets', 1000,
                     'Number of quantiles to bucketize feature values into.')
flags.DEFINE_multi_string('gin_files', [],
                          'List of paths to gin configuration files.')
flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files.')

FLAGS = flags.FLAGS


def _get_feature_info(
    serialized_proto: tf.Tensor,
    features_to_not_process: Iterable[str]) -> Dict[str, tf.io.RaggedFeature]:
  """Provides feature information by analyzing a single serialized example.

  Args:
    serialized_proto: serialized SequenceExample.
    features_to_not_process: A list of feature names that should not be
                             processed.

  Returns:
    Dictionary of Tensor formats indexed by feature name.
  """
  example = tf.train.SequenceExample()
  example.ParseFromString(serialized_proto.numpy())
  sequence_features = {}
  for key, feature_list in example.feature_lists.feature_list.items():
    if key in features_to_not_process:
      continue
    feature = feature_list.feature[0]
    kind = feature.WhichOneof('kind')
    if kind == 'float_list':
      sequence_features[key] = tf.io.RaggedFeature(
          partitions=(), dtype=tf.float32)
    elif kind == 'int64_list':
      sequence_features[key] = tf.io.RaggedFeature(
          partitions=(), dtype=tf.int64)
  return sequence_features


def create_tfrecord_parser_fn(
    sequence_features: Dict[str, tf.io.RaggedFeature]
) -> Callable[[str], List[tf.Tensor]]:
  """Create a parser function for reading serialized tf.data.TFRecordDataset.

  Args:
    sequence_features: Dict having feature type info indexed by feature name.

  Returns:
    A callable that takes scalar serialized proto Tensors
    and emits a list of parsed Tensors for all features.
  """

  def _parser_fn(serialized_proto):
    """Helper function that is returned by create_tfrecord_parser_fn."""
    context_features = {}

    # pylint: enable=g-complex-comprehension
    with tf.name_scope('parse'):
      try:
        _, parsed_sequence = tf.io.parse_single_sequence_example(
            serialized_proto,
            context_features=context_features,
            sequence_features=sequence_features)

        parsed_array = []
        for feature_name in sorted(sequence_features):
          v = parsed_sequence[feature_name]
          if isinstance(v, tf.RaggedTensor):
            v = v.to_tensor()
          parsed_array.append(tf.reshape(v, [-1]))
        return parsed_array
      except ValueError as e:
        # ignore malformed or invalid serialized_proto inputs
        print(f'Error: {e}')

  return _parser_fn


def _generate_vocab(feature_values_arrays, feature_name):
  """Downsample and generate vocab using brute force method."""
  feature_values = np.concatenate(feature_values_arrays)
  sample_length = math.floor(
      np.shape(feature_values)[0] * FLAGS.sampling_fraction)
  values = np.random.choice(feature_values, sample_length, replace=False)
  bin_edges = np.quantile(values, np.linspace(0, 1, FLAGS.num_buckets))
  filename = os.path.join(FLAGS.output_dir, f'{feature_name}.buckets')
  with open(filename, 'w', encoding='utf-8') as f:
    for edge in bin_edges:
      f.write(f'{edge}\n')


def main(_) -> None:
  gin.parse_config_files_and_bindings(
      FLAGS.gin_files, bindings=FLAGS.gin_bindings, skip_unknown=False)
  logging.info(gin.config_str())
  problem_config = registry.get_configuration()

  # Generate num_buckets quantiles for each feature.
  tf.io.gfile.makedirs(FLAGS.output_dir)
  dataset = tf.data.Dataset.list_files(FLAGS.input)
  dataset = tf.data.TFRecordDataset(dataset)
  features_to_not_process = problem_config.get_nonnormalized_features()

  sequence_features = {}
  # TODO(b/222775595): need to fix this after update to logic for handling
  # empty examples during trace generation.
  for raw_example in dataset:
    try:
      sequence_features = _get_feature_info(raw_example,
                                            features_to_not_process)
      logging.info('Found valid sequence_features dict: %s', sequence_features)
      break
    except IndexError:
      # modules with no results have empty feature values and
      # raise an IndexError. For example, in the inliner case, maybe there were
      # no inlining opportunities (very small modules)
      # continue until a module with non-empty feature values is found.
      logging.warn('Found module with empty feature values.')
      continue
  if not sequence_features:
    raise ValueError('No module with non-empty sequence_features values found.')

  parser_fn = create_tfrecord_parser_fn(sequence_features)
  dataset = dataset.map(parser_fn, num_parallel_calls=tf.data.AUTOTUNE)
  data_list = np.array(list(dataset.as_numpy_iterator()), dtype=object)
  data_list = data_list.swapaxes(0, 1)

  with mp.Pool(FLAGS.parallelism) as pool:
    feature_names = list(sorted(sequence_features))
    for i, feature_values_arrays in enumerate(data_list):
      pool.apply_async(_generate_vocab, (
          feature_values_arrays,
          feature_names[i],
      ))
    pool.close()
    pool.join()


if __name__ == '__main__':
  flags.mark_flag_as_required('input')
  flags.mark_flag_as_required('output_dir')
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for collect data of regalloc priority prediction."""

import gin
import tensorflow as tf

import os
import tempfile
from typing import Dict, Optional, Tuple

from compiler_opt.rl import compilation_runner
from compiler_opt.rl import log_reader


@gin.configurable(module='runners')
class RegAllocPriorityRunner(compilation_runner.CompilationRunner):
  """Class for collecting data for regalloc-priority-prediction."""

  def _compile_fn(
      self, file_paths: Tuple[str, ...], tf_policy_path: str, reward_only: bool,
      workdir: str, cancellation_manager: Optional[
          compilation_runner.WorkerCancellationManager]
  ) -> Dict[str, Tuple[tf.train.SequenceExample, float]]:

    file_paths = file_paths[0].replace('.bc', '')
    working_dir = tempfile.mkdtemp(dir=workdir)

    log_path = os.path.join(working_dir, 'log')
    output_native_path = os.path.join(working_dir, 'native')

    result = {}
    command_line = []
    if self._launcher_path:
      command_line.append(self._launcher_path)
    command_line.extend([self._clang_path] + [
        '-c', file_paths, '-O3', '-mllvm', '-regalloc-priority-training-log=' +
        log_path, '-mllvm', '-regalloc-enable-priority-advisor=development',
        '-o', output_native_path
    ])

    if tf_policy_path:
      command_line.extend(
          ['-mllvm', '-regalloc-priority-model=' + tf_policy_path])
    compilation_runner.start_cancellable_process(command_line,
                                                 self._compilation_timeout,
                                                 cancellation_manager)

    # TODO(#202)
    log_result = log_reader.read_log_as_sequence_examples(log_path)

    for fct_name, trajectory in log_result.items():
      if not trajectory.HasField('feature_lists'):
        continue
      score = (
          trajectory.feature_lists.feature_list['reward'].feature[-1].float_list
          .value[0])
      if reward_only:
        result[fct_name] = (None, score)
      else:
        del trajectory.feature_lists.feature_list['reward']
        result[fct_name] = (trajectory, score)

    return result



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Constants for policy training."""

import dataclasses
import enum
import gin
import json

BASE_DIR = 'compiler_opt/rl'
BASE_MODULE_DIR = 'compiler_opt.rl'

# Delta to add when computing reward.
DELTA = 0.01

# Default of global_command_override in corpus_description.json
UNSPECIFIED_OVERRIDE = ['<UNSPECIFIED>']


@gin.constants_from_enum
class AgentName(enum.Enum):
  """Class that enumerates different types of agent names."""
  BEHAVIORAL_CLONE = 0
  DQN = 1
  PPO = 2
  PPO_DISTRIBUTED = 3


class DataClassJSONEncoder(json.JSONEncoder):

  def default(self, o):
    if dataclasses.is_dataclass(o):
      return dataclasses.asdict(o)
    return super().default(o)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A module that allows for easily graphing feature importance data in
notebooks"""

import numpy
import numpy.typing
import shap
import json

from typing import Dict, List, Union, Optional

DataType = Dict[str, Union[numpy.typing.ArrayLike, List[str]]]


def load_shap_values(file_name: str) -> DataType:
  """Loads a set of shap values created with the feature_importance.py script
  into a dictionary that can then be used for creating graphs

  Args:
    file_name: The name of the file in which the shap values are stored. What
      the --output_path flag was set to in the feature importance script.
  """
  with open(file_name, encoding='utf-8') as file_to_load:
    data = json.load(file_to_load)
    if data['expected_values'] is not list:
      data['expected_values'] = [data['expected_values']]
    return {
        'expected_values': numpy.asarray(data['expected_values']),
        'shap_values': numpy.asarray(data['shap_values']),
        'data': numpy.asarray(data['data']),
        'feature_names': data['feature_names']
    }


def init_shap_for_notebook():
  """Initalizes some JS code for interactive feature importance plots."""
  shap.initjs()


def graph_individual_example(data: DataType, index: Optional[int]):
  """Creates a force plot for an example

  Args:
    data: An object containing all the shap values and other information
      necessary to create the plot. Should be created with load_shap_values.
    index: The index of the example that you wish to plot.
  """
  return shap.force_plot(
      data['expected_values'],
      data['shap_values'][index, :],
      data['data'][index, :],
      feature_names=data['feature_names'])


def graph_summary_plot(data: DataType):
  """Creates a summary plot of the entire dataset given

  Args:
    data: An object containing all the shap values necessary to create the
      plot. Should come from load_shap_values
  """
  return shap.summary_plot(
      data['shap_values'], data['data'], feature_names=data['feature_names'])



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for compiler_opt.rl.data_reader."""

# pylint: disable=protected-access

import os

from absl.testing import parameterized
import tensorflow as tf
from tf_agents.specs import tensor_spec
from tf_agents.trajectories import time_step
from tf_agents.trajectories import trajectory

from compiler_opt.rl import agent_config
from compiler_opt.rl import data_reader


def _define_sequence_example(agent_config_type, is_action_discrete):
  example = tf.train.SequenceExample()
  for _ in range(10):
    example.feature_lists.feature_list['feature_key'].feature.add(
    ).int64_list.value.append(1)
    if is_action_discrete:
      example.feature_lists.feature_list['inlining_decision'].feature.add(
      ).int64_list.value.append(0)
    else:
      example.feature_lists.feature_list['live_interval_weight'].feature.add(
      ).float_list.value.append(1.23)
    example.feature_lists.feature_list['reward'].feature.add(
    ).float_list.value.append(2.3)
    if agent_config_type in (agent_config.PPOAgentConfig,
                             agent_config.DistributedPPOAgentConfig):
      if is_action_discrete:
        example.feature_lists.feature_list[
            'CategoricalProjectionNetwork_logits'].feature.add(
            ).float_list.value.extend([1.2, 3.4])
      else:
        example.feature_lists.feature_list[
            'NormalProjectionNetwork_scale'].feature.add(
            ).float_list.value.extend([1.2])
        example.feature_lists.feature_list[
            'NormalProjectionNetwork_loc'].feature.add(
            ).float_list.value.extend([3.4])
  return example


def _write_tmp_tfrecord(file_path, example, num_examples):
  with tf.io.TFRecordWriter(file_path) as file_writer:
    for _ in range(num_examples):
      file_writer.write(example.SerializeToString())


class DataReaderTest(tf.test.TestCase, parameterized.TestCase):

  def setUp(self):
    observation_spec = {
        'feature_key':
            tf.TensorSpec(dtype=tf.int64, shape=(), name='feature_key')
    }
    reward_spec = tf.TensorSpec(dtype=tf.float32, shape=(), name='reward')
    self._time_step_spec = time_step.time_step_spec(observation_spec,
                                                    reward_spec)
    self._discrete_action_spec = tensor_spec.BoundedTensorSpec(
        dtype=tf.int64,
        shape=(),
        minimum=0,
        maximum=1,
        name='inlining_decision')
    self._continuous_action_spec = tensor_spec.BoundedTensorSpec(
        dtype=tf.float32,
        shape=(),
        name='live_interval_weight',
        minimum=-100,
        maximum=20)
    super().setUp()

  def _create_sequence_example_datasource(self, example):
    return [example.SerializeToString()] * 100

  def _create_tfrecord_datasource(self, example):
    data_source = os.path.join(self.get_temp_dir(), 'data_tfrecord')
    _write_tmp_tfrecord(data_source, example, 100)
    return data_source

  _test_config = (('SequenceExampleDatasetFn',
                   data_reader.create_sequence_example_dataset_fn,
                   agent_config.PPOAgentConfig,
                   _create_sequence_example_datasource),
                  ('TFRecordDatasetFn', data_reader.create_tfrecord_dataset_fn,
                   agent_config.PPOAgentConfig, _create_tfrecord_datasource))

  @parameterized.named_parameters(*_test_config)
  def test_create_dataset_fn(self, test_fn, _, data_source_fn):
    agent_type_override = agent_config.DQNAgentConfig
    example = _define_sequence_example(
        agent_type_override, is_action_discrete=True)

    data_source = data_source_fn(self, example)
    dataset_fn = test_fn(
        agent_cfg=agent_type_override(
            time_step_spec=self._time_step_spec,
            action_spec=self._discrete_action_spec),
        batch_size=2,
        train_sequence_length=3)
    data_iterator = iter(dataset_fn(data_source).repeat())

    experience = next(data_iterator)
    self.assertIsInstance(experience, trajectory.Trajectory)
    self.assertAllEqual([2, 3], experience.step_type.shape)
    self.assertCountEqual(['feature_key'], experience.observation.keys())
    self.assertAllEqual([[1, 1, 1], [1, 1, 1]],
                        experience.observation['feature_key'])
    self.assertAllEqual([[0, 0, 0], [0, 0, 0]], experience.action)
    self.assertEmpty(experience.policy_info)
    self.assertAllEqual([2, 3], experience.next_step_type.shape)
    self.assertAllClose([[2.3, 2.3, 2.3], [2.3, 2.3, 2.3]], experience.reward)
    self.assertAllEqual([[1, 1, 1], [1, 1, 1]], experience.discount)

  _distrib_test_config = (('SequenceExampleDatasetFnDistributed',
                           data_reader.create_sequence_example_dataset_fn,
                           agent_config.DistributedPPOAgentConfig,
                           _create_sequence_example_datasource),
                          ('TFRecordDatasetFnDistributed',
                           data_reader.create_tfrecord_dataset_fn,
                           agent_config.DistributedPPOAgentConfig,
                           _create_tfrecord_datasource))

  @parameterized.named_parameters(*(_test_config + _distrib_test_config))
  def test_ppo_policy_info_discrete(self, test_fn, agent_config_type,
                                    data_source_fn):
    example = _define_sequence_example(
        agent_config_type, is_action_discrete=True)

    data_source = data_source_fn(self, example)

    dataset_fn = test_fn(
        agent_cfg=agent_config_type(
            time_step_spec=self._time_step_spec,
            action_spec=self._discrete_action_spec),
        batch_size=2,
        train_sequence_length=3)
    data_iterator = iter(dataset_fn(data_source).repeat())

    experience = next(data_iterator)
    self.assertAllEqual(['feature_key'], list(experience.observation.keys()))
    self.assertAllClose([[[1.2, 3.4], [1.2, 3.4], [1.2, 3.4]],
                         [[1.2, 3.4], [1.2, 3.4], [1.2, 3.4]]],
                        experience.policy_info['dist_params']['logits'])

  @parameterized.named_parameters(*(_test_config + _distrib_test_config))
  def test_ppo_policy_info_continuous(self, test_fn, agent_config_type,
                                      data_source_fn):
    example = _define_sequence_example(
        agent_config_type, is_action_discrete=False)

    data_source = data_source_fn(self, example)

    dataset_fn = test_fn(
        agent_cfg=agent_config_type(
            time_step_spec=self._time_step_spec,
            action_spec=self._continuous_action_spec),
        batch_size=2,
        train_sequence_length=3)
    data_iterator = iter(dataset_fn(data_source).repeat())

    experience = next(data_iterator)
    self.assertAllEqual(['feature_key'], list(experience.observation.keys()))
    self.assertAllClose([[1.23, 1.23, 1.23], [1.23, 1.23, 1.23]],
                        experience.action)
    self.assertAllClose([[1.2, 1.2, 1.2], [1.2, 1.2, 1.2]],
                        experience.policy_info['dist_params']['scale'])
    self.assertAllClose([[3.4, 3.4, 3.4], [3.4, 3.4, 3.4]],
                        experience.policy_info['dist_params']['loc'])


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for collect data of regalloc-for-performance."""

import os
import tempfile
from typing import Dict, Tuple

import gin
import tensorflow as tf

from compiler_opt.rl import compilation_runner
from compiler_opt.rl import corpus
from compiler_opt.rl import log_reader


@gin.configurable(module='runners')
class RegAllocRunner(compilation_runner.CompilationRunner):
  """Class for collecting data for regalloc-for-performance.

  Usage:
  runner = RegAllocRunner(
               clang_path, launcher_path, moving_average_decay_rate)
  serialized_sequence_example, default_reward, moving_average_reward,
  policy_reward = runner.collect_data(
      ir_path, tf_policy_path, default_reward, moving_average_reward)
  """

  # TODO: refactor file_paths parameter to ensure correctness during
  # construction
  def compile_fn(
      self, command_line: corpus.FullyQualifiedCmdLine, tf_policy_path: str,
      reward_only: bool,
      workdir: str) -> Dict[str, Tuple[tf.train.SequenceExample, float]]:
    """Run the compiler for the given IR file under the given policy.

    Args:
      command_line: the fully qualified command line.
      tf_policy_path: path to TF policy direcoty on local disk.
      reward_only: whether only return reward.

    Returns:
      A dict mapping from example identifier to tuple containing:
        sequence_example: A tf.SequenceExample proto describing compilation
          trace, None if reward_only == True.
        reward: reward of register allocation.

    Raises:
      subprocess.CalledProcessError: if process fails.
      compilation_runner.ProcessKilledError: (which it must pass through) on
        cancelled work.
      RuntimeError: if llvm-size produces unexpected output.
    """

    working_dir = tempfile.mkdtemp(dir=workdir)

    log_path = os.path.join(working_dir, 'log')
    output_native_path = os.path.join(working_dir, 'native')

    result = {}
    cmdline = []
    if self._launcher_path:
      cmdline.append(self._launcher_path)
    cmdline.extend([self._clang_path] + list(command_line) + [
        '-mllvm', '-regalloc-enable-advisor=development', '-mllvm',
        '-regalloc-training-log=' + log_path, '-o', output_native_path
    ])

    if tf_policy_path:
      cmdline.extend(['-mllvm', '-regalloc-model=' + tf_policy_path])
    compilation_runner.start_cancellable_process(cmdline,
                                                 self._compilation_timeout,
                                                 self._cancellation_manager)

    # TODO(#202)
    log_result = log_reader.read_log_as_sequence_examples(log_path)

    for fct_name, trajectory in log_result.items():
      if not trajectory.HasField('feature_lists'):
        continue
      score = (
          trajectory.feature_lists.feature_list['reward'].feature[-1].float_list
          .value[0])
      if reward_only:
        result[fct_name] = (None, score)
      else:
        del trajectory.feature_lists.feature_list['reward']
        result[fct_name] = (trajectory, score)

    return result



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A script for running chromium based benchmarks

This script allows for running chromium benchmarks to evaluate the performance
of MLGO regalloc models in a highly automated fashion. It will automatically
recompile LLVM if requested with the release mode model specified, recompile
the chromium benchmarks using the correct MLGO model/advisor, and then run
a specified subset of those tests designed to minimize run to run variability.

Usage:
PYTHONPATH=$PYTHONPATH:. python3 \
  ./compiler_opt/benchmark/benchmark_chromium.py \
  --compile_tests \
  --advisor=release \
  --chromium_src_path=/chromium/src \
  --depot_tools_path=/depot_tools \
  --llvm_build_path=/llvm-build \
  --compile_llvm \
  --model_path=/tmp/model \
  --chromium_build_path=./out/Release \
  --output_file=./output.json \
  --perf_counters=mem_uops_retired.all_loads \
  --perf_counters=mem_uops_retired.all_stores

Note that --perf_counters can be defined multiple times to grab more than one
performance counter. Also note that the chromium_build_path is a relative
directory. It is relative to the chromium source dir and specified this way
as there appears to be problems building chromium outside of the source
directory.
"""

import os
import shutil
import subprocess
import json

from absl import flags
from absl import app

from compiler_opt.benchmark import gtest_executable_utils
from compiler_opt.benchmark import benchmarking_utils

from typing import List, Dict, Union

FLAGS = flags.FLAGS

test_prefix = './compiler_opt/benchmark/chromium_test_descriptions/'

test_description_files = [
    'base_perftests.json', 'browser_tests.json', 'components_perftests.json',
    'base_unittests.json', 'cc_unittests.json', 'components_unittests.json',
    'content_unittests.json'
]

default_test_descriptions = [
    f'{test_prefix}{test_dsc}' for test_dsc in test_description_files
]

flags.DEFINE_multi_string(
    'test_description', default_test_descriptions,
    '(Can be defined multiple times) A path to a test'
    'description JSON file containing the test executable'
    'and the tests to run')
flags.DEFINE_boolean('compile_tests', True,
                     'Whether or not to compile the tests from scratch')
flags.DEFINE_enum('advisor', None, ['release', 'default'],
                  'The advisor to use when compiling chromium')
flags.DEFINE_string('chromium_src_path', '/chromium/src',
                    'The path to the chromium source')
flags.DEFINE_string('depot_tools_path', '/depot_tools',
                    'The path to your depot tools checkout')
flags.DEFINE_string('llvm_build_path', '/llvm-build',
                    'The path to your llvm build')
flags.DEFINE_boolean('compile_llvm', True,
                     'whether or not to compile llvm using the new model')
flags.DEFINE_boolean(
    'llvm_use_incremental', True,
    'whether or not to use an incremental build while'
    'compiling llvm')
flags.DEFINE_string('llvm_source_path', '/llvm-project',
                    'The root path of your local llvm-project checkout')
flags.DEFINE_string('model_path', '',
                    'The path to the model to use when compiling llvm')
flags.DEFINE_string(
    'chromium_build_path', './out/Release',
    'The chromium build path, relative to the chromium source'
    'directory')
flags.DEFINE_string('output_file', 'output.json',
                    'The path to the output file (in JSON format)')
flags.DEFINE_integer(
    'num_threads', 1, 'The number of threads to use when running benchmarks.'
    'Should be used with caution')
flags.DEFINE_multi_string(
    'perf_counters',
    ['mem_uops_retired.all_loads', 'mem_uops_retired.all_stores'],
    'The performance counters to use')


def build_chromium_tests(regalloc_advisor: str, chromium_build_path: str,
                         chromium_source_path: str, depot_tools_path: str,
                         llvm_build_path: str, tests_to_build: List[str]):
  """Builds the chromium test suite

  This function will build the specified chromium tests using the specified
  regalloc advisor. This function configures some default build options using gn
  as shown below in the gn_args list, and then builds the needed targets using
  autoninja.

  Args:
    regalloc_advisor: The regalloc advisor to use when compiling
    chromium_build_path: The path (relative to the chromium source dir) to use
      for building chromium
    chromium_source_path: The path to the chromium source
    depot_tools_path: The path to the root of your depot tools checkout
    llvm_build_path: The path to the root of the directory where llvm was built
    tests_to_build: An array of test targets that are to be built
  """
  chromium_absolute_build_path = os.path.join(chromium_source_path,
                                              chromium_build_path)
  if os.path.exists(chromium_absolute_build_path):
    shutil.rmtree(chromium_absolute_build_path)

  new_environment = os.environ.copy()
  new_environment['PATH'] += ':' + depot_tools_path
  new_environment['CC'] = os.path.join(llvm_build_path, './bin/clang')
  new_environment['CXX'] = os.path.join(llvm_build_path, './bin/clang++')
  new_environment['AR'] = os.path.join(llvm_build_path, './bin/llvm-ar')
  new_environment['NM'] = os.path.join(llvm_build_path, './bin/llvm-nm')
  new_environment['CPPFLAGS'] = \
    f'-mllvm -regalloc-enable-advisor={regalloc_advisor}'

  gn_args = [
      'is_official_build=true', 'use_thin_lto=false', 'is_cfi=false',
      'use_cfi_icall=false', 'use_cfi_cast=false',
      'clang_use_chrome_plugins=false', 'is_debug=false', 'symbol_level=0',
      'custom_toolchain=\\\"//build/toolchain/linux/unbundle:default\\\"',
      'host_toolchain=\\\"//build/toolchain/linux/unbundle:default\\\"'
  ]

  gn_args_string = '--args="'
  for arg in gn_args:
    gn_args_string += arg + ' '
  gn_args_string += '"'

  gn_config_command = 'gn gen ' + chromium_build_path + ' ' + gn_args_string
  with subprocess.Popen(
      gn_config_command,
      env=new_environment,
      cwd=chromium_source_path,
      shell=True) as gn_config_process:
    gn_config_process.wait()

  ninja_compile_command = ['autoninja', '-C', chromium_build_path]
  ninja_compile_command.extend(tests_to_build)
  with subprocess.Popen(
      ninja_compile_command, env=new_environment,
      cwd=chromium_source_path) as ninja_compile_process:
    ninja_compile_process.wait()


def run_tests(tests_to_run: List[Dict[str, Union[str, List[str]]]],
              chromium_absolute_build_path: str, num_threads: int,
              perf_counters: List[str]):
  """A utility to run a set of chromium tests

  This function takes in a list of test descriptions containing the
  name of a chromium test target as well as a list of all the tests
  within that test executable that are to be run. It executes each test
  (in parallel if specified for performance counters that aren't highly
  sensitive to that environment), grabbing the perf counters for the test
  that are specified.

  Args:
    tests_to_run: A list of python dictionaries containing the test descriptions
    chromium_absolute_build_path: The absolute build path to the chromium
      build dir
    num_threads: The number of threads to use when running tests
    perf_counters: A list of perf compatible performance counters
  """
  test_data = []
  for test in tests_to_run:
    executable_path = os.path.join(chromium_absolute_build_path,
                                   test['executable'])
    test_data.extend(
        gtest_executable_utils.run_test_suite(test, executable_path,
                                              perf_counters, num_threads))
  return test_data


def main(_):
  test_descriptions = []
  for test_description in FLAGS.test_description:
    with open(test_description, encoding='UTF-8') as test_description_file:
      print(test_description)
      test_descriptions.append(json.load(test_description_file))
  test_executables = []
  for test_description in test_descriptions:
    test_executables.append(test_description['executable'])

  if FLAGS.compile_llvm:
    benchmarking_utils.build_llvm(FLAGS.model_path, FLAGS.llvm_use_incremental,
                                  FLAGS.llvm_build_path, FLAGS.llvm_source_path)

  if FLAGS.compile_tests:
    build_chromium_tests(FLAGS.advisor, FLAGS.chromium_build_path,
                         FLAGS.chromium_src_path, FLAGS.depot_tools_path,
                         FLAGS.llvm_build_path, test_executables)

  chromium_absolute_build_path = os.path.join(FLAGS.chromium_src_path,
                                              FLAGS.chromium_build_path)
  test_data = run_tests(test_descriptions, chromium_absolute_build_path,
                        FLAGS.num_threads, FLAGS.perf_counters)

  with open(FLAGS.output_file, 'w', encoding='UTF-8') as output_file:
    output_data = {'benchmarks': test_data}
    output_file.write(json.dumps(output_data, indent=4))


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Inlining Training config."""

import gin
import tensorflow as tf
from tf_agents.specs import tensor_spec
from tf_agents.trajectories import time_step
from compiler_opt.rl import feature_ops


# pylint: disable=g-complex-comprehension
@gin.configurable()
def get_inlining_signature_spec():
  """Returns (time_step_spec, action_spec) for LLVM inlining."""
  observation_spec = dict(
      (key, tf.TensorSpec(dtype=tf.int64, shape=(), name=key)) for key in (
          # Base features
          'caller_basic_block_count',
          'caller_conditionally_executed_blocks',
          'caller_users',
          'callee_basic_block_count',
          'callee_conditionally_executed_blocks',
          'callee_users',
          'nr_ctant_params',
          'node_count',
          'edge_count',
          'callsite_height',
          'cost_estimate',

          # Expanded cost features
          'sroa_savings',
          'sroa_losses',
          'load_elimination',
          'call_penalty',
          'call_argument_setup',
          'load_relative_intrinsic',
          'lowered_call_arg_setup',
          'indirect_call_penalty',
          'jump_table_penalty',
          'case_cluster_penalty',
          'switch_penalty',
          'unsimplified_common_instructions',
          'num_loops',
          'dead_blocks',
          'simplified_instructions',
          'constant_args',
          'constant_offset_ptr_args',
          'callsite_cost',
          'cold_cc_penalty',
          'last_call_to_static_bonus',
          'is_multiple_blocks',
          'nested_inlines',
          'nested_inline_cost_estimate',
          'threshold',

          # inlining_default is not used as feature in training.
          'inlining_default'))
  reward_spec = tf.TensorSpec(dtype=tf.float32, shape=(), name='reward')
  time_step_spec = time_step.time_step_spec(observation_spec, reward_spec)
  action_spec = tensor_spec.BoundedTensorSpec(
      dtype=tf.int64, shape=(), name='inlining_decision', minimum=0, maximum=1)

  return time_step_spec, action_spec


@gin.configurable
def get_observation_processing_layer_creator(quantile_file_dir=None,
                                             with_sqrt=True,
                                             with_z_score_normalization=True,
                                             eps=1e-8):
  """Wrapper for observation_processing_layer."""
  quantile_map = feature_ops.build_quantile_map(quantile_file_dir)

  def observation_processing_layer(obs_spec):
    """Creates the layer to process observation given obs_spec."""
    if obs_spec.name == 'inlining_default':
      return tf.keras.layers.Lambda(feature_ops.discard_fn)

    quantile = quantile_map[obs_spec.name]
    return tf.keras.layers.Lambda(
        feature_ops.get_normalize_fn(quantile, with_sqrt,
                                     with_z_score_normalization, eps))

  return observation_processing_layer


def get_nonnormalized_features():
  return ['reward', 'inlining_default', 'inlining_decision']



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library to launch a stand alone Reverb RB server."""

import os

from absl import logging

import reverb
import gin
import tensorflow.compat.v2 as tf

from tf_agents.experimental.distributed import reverb_variable_container
from tf_agents.specs import tensor_spec
from tf_agents.train import learner
from tf_agents.train.utils import train_utils
from tf_agents.utils import common


@gin.configurable
def run_reverb_server(root_dir: str,
                      port: int,
                      replay_buffer_capacity: int,
                      num_epochs: int = 5):
  """Start the server after the initial policy becomes available."""
  # Wait for the collect policy to become available, then load it.
  collect_policy_dir = os.path.join(root_dir, learner.POLICY_SAVED_MODEL_DIR,
                                    learner.COLLECT_POLICY_SAVED_MODEL_DIR)
  collect_policy = train_utils.wait_for_policy(
      collect_policy_dir, load_specs_from_pbtxt=True)

  # Create the signature for the variable container holding the policy weights.
  train_step = train_utils.create_train_step()
  model_id = common.create_variable('model_id')
  variables = {
      reverb_variable_container.POLICY_KEY: collect_policy.variables(),
      reverb_variable_container.TRAIN_STEP_KEY: train_step,
      'model_id': model_id,
  }
  variable_container_signature = tf.nest.map_structure(
      lambda variable: tf.TensorSpec(variable.shape, dtype=variable.dtype),
      variables)
  logging.info('Signature of variables: \n%s', variable_container_signature)

  # Create the signature for the replay buffer holding observed experience.
  replay_buffer_signature = tensor_spec.from_spec(
      collect_policy.collect_data_spec)
  replay_buffer_signature = tensor_spec.add_outer_dim(replay_buffer_signature)
  logging.info('Signature of experience: \n%s', replay_buffer_signature)

  # Create and start the replay buffer and variable container server.
  server = reverb.Server(
      tables=[
          # The remover does not matter because we clear the table and the end
          # of each global step. We assume that the table is large enough to
          # contain the data collected from one step.
          reverb.Table(
              name='training_table',
              sampler=reverb.selectors.Fifo(),
              remover=reverb.selectors.Fifo(),
              rate_limiter=reverb.rate_limiters.MinSize(1),
              max_size=replay_buffer_capacity,
              max_times_sampled=num_epochs,
              signature=replay_buffer_signature,
          ),
          reverb.Table(
              name=reverb_variable_container.DEFAULT_TABLE,
              sampler=reverb.selectors.Fifo(),
              remover=reverb.selectors.Fifo(),
              rate_limiter=reverb.rate_limiters.MinSize(1),
              max_size=1,
              max_times_sampled=0,
              signature=variable_container_signature,
          ),
      ],
      port=port)
  server.wait()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Combine multiple training corpus into a single training corpus.

Currently only support the case that multiple corpus share the same
configurables except the "modules" field.

Usage: we'd like to combine training corpus corpus1 and corpus2 into
combinedcorpus; we first structure the files as follows:

combinedcorpus
combinedcorpus/corpus1
combinedcorpus/corpus2

Running this script with

python3 \
compiler_opt/tools/combine_training_corpus.py \
  --root_dir=$PATH_TO_combinedcorpus

generates combinedcorpus/corpus_description.json file. In this way corpus1
and corpus2 are combined into combinedcorpus.
"""

from absl import app
from absl import flags

from compiler_opt.tools import combine_training_corpus_lib

flags.DEFINE_string('root_dir', '', 'root dir of module paths to combine.')

FLAGS = flags.FLAGS


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')


combine_training_corpus_lib.combine_corpus(FLAGS.root_dir)

if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Compilation problem component model.

A 'compilation problem' is an optimization problem with a specific way of
invoking clang and specific features and tensorflow topologies. The component
model requires all these be exported in a class implementing
ProblemConfiguration below, however, to avoid cycle dependencies in Bazel
environments, do not explicitly inherit from it.

Internally, all the module's implementation parameters are expected to be
gin-initialized.

To use:

* just get a ProblemConfiguration object in your python:

`config = config.get_configuration()`

* when running the tool, pass:

  `--gin_bindings=config_registry.get_configuration.implementation=\
    @the.implementation.name`

  for example:

  `--gin_bindings=config_registry.get_configuration.implementation=\
    @configs.InliningConfig`

=================
Conventions
=================

* to avoid long binding names, use the 'runners' module name for the
  CompilationRunner implementation, and use the 'configs' module name for the
  implementation of ProblemConfiguration.

* the CompilationRunner gin initialization should initialize to None, and use,
  the 'clang_path' and 'launcher_path' macros
  (https://github.com/google/gin-config#syntax-quick-reference):

  clang_path = None
  launcher_path = None
  runners.MyCompilationRunner.clang_path = %clang_path
  runners.MyCompilationRunner.launcher_path = %launcher_path

  Use a similar pattern for problem-specific additional flags (see 'inlining'
  and 'llvm_size_path' for example).

  When running tools, this allows the user pass common flags transparently wrt
  the underlying runner - i.e. if swapping 2 runners, the clang flag stays the
  same:

  `--gin_bindings=clang_path="'/foo/bar/clang'"`

"""

import abc
import gin
from typing import Callable, Dict, Iterable, Optional, Tuple

import tensorflow as tf
import tf_agents as tfa

# used for type annotation in a string (for 3.8 compat)
# pylint: disable=unused-import
from compiler_opt.rl import compilation_runner
from compiler_opt.rl import env

types = tfa.typing.types


class ProblemConfiguration(metaclass=abc.ABCMeta):
  """Abstraction of the APIs accessing a problem-specific configuration."""

  @abc.abstractmethod
  def get_env(self) -> env.MLGOEnvironmentBase:
    raise NotImplementedError

  @abc.abstractmethod
  def get_signature_spec(
      self) -> Tuple[types.NestedTensorSpec, types.NestedTensorSpec]:
    raise NotImplementedError

  @abc.abstractmethod
  def get_preprocessing_layer_creator(
      self) -> Callable[[types.TensorSpec], tf.keras.layers.Layer]:
    raise NotImplementedError

  def get_nonnormalized_features(self) -> Iterable[str]:
    return []

  @abc.abstractmethod
  def get_runner_type(self) -> 'type[compilation_runner.CompilationRunner]':
    raise NotImplementedError

  # TODO(b/233935329): The following clang flags need to be tied to a corpus
  # rather than to a training tool invocation.

  # List of flags to add to clang compilation command.
  @gin.configurable(module='problem_config')
  def flags_to_add(self, add_flags=()) -> Tuple[str, ...]:
    return add_flags

  # List of flags to remove from clang compilation command. The flag names
  # should match the actual flags provided to clang.'
  @gin.configurable(module='problem_config')
  def flags_to_delete(self, delete_flags=()) -> Tuple[str, ...]:
    return delete_flags

  # List of flags to replace in the clang compilation command. The flag names
  # should match the actual flags provided to clang. An example for AFDO
  # reinjection:
  # replace_flags={
  #     '-fprofile-sample-use':'/path/to/gwp.afdo',
  #     '-fprofile-remapping-file':'/path/to/prof_remap.txt'
  # }
  # return replace_flags
  @gin.configurable(module='problem_config')
  def flags_to_replace(self, replace_flags=None) -> Optional[Dict[str, str]]:
    return replace_flags



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declaration of a few external GIN configurable functions."""

from gin import config

# pylint: disable=unused-import
from gin.tf import external_configurables
# pylint: enable=unused-import

import tensorflow as tf

# Activation function.
config.external_configurable(tf.keras.activations.relu,
                             'tf.keras.activations.relu')

# Layers
config.external_configurable(tf.keras.layers.Add, 'tf.keras.layers.Add')
config.external_configurable(tf.keras.layers.Concatenate,
                             'tf.keras.layers.Concatenate')

# Loss functions
config.external_configurable(tf.keras.losses.MeanSquaredError,
                             'tf.keras.losses.MeanSquaredError')

config.external_configurable(tf.keras.losses.CategoricalCrossentropy,
                             'tf.keras.losses.CategoricalCrossentropy')



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for random_network_distillation."""

from absl.testing import parameterized
import tensorflow as tf
from tf_agents.networks import encoding_network
from tf_agents.networks import expand_dims_layer
from tf_agents.trajectories import time_step
from tf_agents.trajectories import trajectory

from compiler_opt.rl import random_net_distillation


def _processing_layer_creator():
  expand_dims_op = expand_dims_layer.ExpandDims(-1)

  def observation_processing_layer(_):
    """Creates the toy layer to process observation."""

    def discard_feature(obs):
      expanded_obs = expand_dims_op(obs)
      return tf.ones_like(expanded_obs, dtype=tf.float32)

    func = discard_feature
    return tf.keras.layers.Lambda(func)

  return observation_processing_layer


def _create_test_data(batch_size, sequence_length):
  test_trajectory = trajectory.Trajectory(
      step_type=tf.fill([batch_size, sequence_length], 1),
      observation={
          'edge_count':
              tf.fill([batch_size, sequence_length],
                      tf.constant(10, dtype=tf.int64))
      },
      action=tf.fill([batch_size, sequence_length],
                     tf.constant(1, dtype=tf.int64)),
      policy_info=(),
      next_step_type=tf.fill([batch_size, sequence_length], 1),
      reward=tf.fill([batch_size, sequence_length], 2.0),
      discount=tf.fill([batch_size, sequence_length], 1.0),
  )

  def test_data_iterator():
    while True:
      yield test_trajectory

  return test_data_iterator()


class RandomNetworkDistillationTest(tf.test.TestCase, parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._update_frequency = 1
    observation_spec = {
        'edge_count': tf.TensorSpec(
            dtype=tf.int64, shape=(), name='edge_count')
    }
    self._time_step_spec = time_step.time_step_spec(observation_spec)

  def test_train_on_batch(self):
    data_iterator = _create_test_data(batch_size=3, sequence_length=3)

    # initialize the random_network_distillation instance
    random_network_distillation = \
        random_net_distillation.RandomNetworkDistillation(
          time_step_spec=self._time_step_spec,
          preprocessing_layer_creator=_processing_layer_creator(),
          encoding_network=encoding_network.EncodingNetwork,
          update_frequency=self._update_frequency)

    experience = next(data_iterator)
    # test the RND train function return type
    for _ in range(5):
      new_experience = random_network_distillation.train(experience)
    self.assertIsInstance(new_experience, trajectory.Trajectory)
    # the rest of experience should remain the same except reward
    self.assertAllEqual(experience.step_type, new_experience.step_type)
    self.assertAllEqual(experience.observation, new_experience.observation)
    self.assertAllEqual(experience.action, new_experience.action)
    self.assertAllEqual(experience.policy_info, new_experience.policy_info)
    self.assertAllEqual(experience.next_step_type, experience.next_step_type)
    self.assertAllEqual(experience.discount, new_experience.discount)
    # reward should have same shape
    self.assertAllEqual(experience.reward.shape, new_experience.reward.shape)
    # new reward should has finite value
    self.assertFalse(tf.math.is_inf(tf.reduce_sum(new_experience.reward)))


if __name__ == '__main__':
  tf.test.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility to create MLGO policy learner."""

from typing import Callable, List, Optional, Text, Tuple

from absl import logging

import tensorflow as tf
from tf_agents.agents.ppo import ppo_agent
from tf_agents.train import interval_trigger
from tf_agents.train import learner
from tf_agents.typing import types

# A function which processes a tuple of a nested tensor representing a TF-Agent
# Trajectory and Reverb SampleInfo.
_SequenceParamsType = Tuple[types.NestedTensor, types.ReverbSampleInfo]
_SequenceFnType = Callable[[_SequenceParamsType], _SequenceParamsType]


class MLGOPPOLearner(object):
  """Manages all the learning details needed.

  These include:
    * Using distribution strategies correctly
    * Summaries
    * Checkpoints
    * Minimizing entering/exiting TF context:
        Especially in the case of TPUs scheduling a single TPU program to
        perform multiple train steps is critical for performance.
    * Generalizes the train call to be done correctly across CPU, GPU, or TPU
      executions managed by DistributionStrategies. This uses `strategy.run` and
      then makes sure to do a reduce operation over the `LossInfo` returned by
      the agent.
  """

  def __init__(self,
               root_dir: Text,
               train_step: tf.Variable,
               model_id: tf.Variable,
               agent: ppo_agent.PPOAgent,
               experience_dataset_fn: Callable[[], tf.data.Dataset],
               sequence_length: int,
               num_episodes_per_iteration: int,
               minibatch_size: int,
               shuffle_buffer_size: int,
               num_epochs: int = 1,
               triggers: Optional[List[
                   interval_trigger.IntervalTrigger]] = None,
               checkpoint_interval: int = 100000,
               summary_interval: int = 1000,
               strategy: Optional[tf.distribute.Strategy] = None,
               per_sequence_fn: Optional[_SequenceFnType] = None,
               allow_variable_length_episodes: bool = False) -> None:
    """Initializes a MLGOPPOLearner instance.

    Args:
      root_dir: Main directory path where checkpoints, saved_models, and
        summaries will be written to.
      train_step: a scalar tf.int64 `tf.Variable` which will keep track of the
        number of train steps. This is used for artifacts created like
        summaries, or outputs in the root_dir.
      model_id: a scalar tf.int64 `tf.Variable` which will keep track of the
        number of learner iterations / policy updates.
      agent: `ppo_agent.PPOAgent` instance to train with. Note that
        update_normalizers_in_train should be set to `False`, otherwise a
        ValueError will be raised. We do not update normalizers in the agent
        again because we already update it in the learner. When mini batching is
        enabled, compute_value_and_advantage_in_train should be set to False,
        and preprocessing should be done as part of the data pipeline as part of
        `replay_buffer.as_dataset`.
      experience_dataset_fn: a function that will create an instance of a
        tf.data.Dataset used to sample experience for training. Each element in
        the dataset is a (Trajectory, SampleInfo) pair.
      sequence_length: Fixed sequence length for elements in the dataset. Used
        for calculating how many iterations of minibatches to use for training.
      num_episodes_per_iteration: The number of episodes to sample for training.
        If fewer than this amount of episodes exists in the dataset, the learner
        will wait for more data to be added, or until the reverb timeout is
        reached.
      minibatch_size: The minibatch size. The dataset used for training is
        shaped `[minibatch_size, 1, ...]`. If None, full sequences will be fed
        into the agent. Please set this parameter to None for RNN networks which
        requires full sequences.
      shuffle_buffer_size: The buffer size for shuffling the trajectories before
        splitting them into mini batches. Only required when mini batch learning
        is enabled (minibatch_size is set). Otherwise it is ignored. Commonly
        set to a number 1-3x the episode length of your environment.
      num_epochs: The number of iterations to go through the same sequences.
      triggers: List of callables of the form `trigger(train_step)`. After every
        `run` call every trigger is called with the current `train_step` value
        as an np scalar.
      checkpoint_interval: Number of train steps in between checkpoints. Note
        these are placed into triggers and so a check to generate a checkpoint
        only occurs after every `run` call. Set to -1 to disable (this is not
        recommended, because it means that if the pipeline gets preempted, all
        previous progress is lost). This only takes care of the checkpointing
        the training process.  Policies must be explicitly exported through
        triggers.
      summary_interval: Number of train steps in between summaries. Note these
        are placed into triggers and so a check to generate a checkpoint only
        occurs after every `run` call.
      strategy: (Optional) `tf.distribute.Strategy` to use during training.
      per_sequence_fn: (Optional): sequence-wise preprecessing, pass in agent.
        preprocess for advantage calculation. This operation happens after
        take() and before rebatching.
      allow_variable_length_episodes: Whether to support variable length
        episodes for training.

    Raises:
      ValueError: agent._compute_value_and_advantage_in_train is set to `True`.
        preprocessing must be done as part of the data pipeline when mini
        batching is enabled.
    """

    self._strategy = strategy or tf.distribute.get_strategy()
    self._agent = agent
    self._minibatch_size = minibatch_size
    self._shuffle_buffer_size = shuffle_buffer_size
    self._num_epochs = num_epochs
    self._experience_dataset_fn = experience_dataset_fn
    self._num_episodes_per_iteration = num_episodes_per_iteration
    # Tracks the number of times learner.run() has been called.
    # This is used for filtering out data generated by older models to ensure
    # the on policyness of the algorithm.
    self._model_id = model_id
    self._sequence_length = sequence_length
    self._per_sequence_fn = per_sequence_fn

    self._generic_learner = learner.Learner(
        root_dir,
        train_step,
        agent,
        after_train_strategy_step_fn=None,
        triggers=triggers,
        checkpoint_interval=checkpoint_interval,
        summary_interval=summary_interval,
        use_kwargs_in_agent_train=False,
        strategy=self._strategy)

    self.num_replicas = self._strategy.num_replicas_in_sync
    self._allow_variable_length_episodes = allow_variable_length_episodes
    self._num_samples = self._num_episodes_per_iteration * self._sequence_length
    self._create_datasets()
    self._steps_per_iter = self._get_train_steps_per_iteration()

    logging.info('[Learner] Replicas for training: %d', self.num_replicas)
    logging.info('[Learner] Samples per iteration: %d', self._num_samples)
    logging.info('[Learner] Train steps per iteration: %d',
                 self._steps_per_iter)

  def _create_datasets(self):
    """Create the training dataset and iterator."""

    def _filter_invalid_episodes(sample):
      sample_info = sample.info
      data_model_id = tf.cast(
          tf.reduce_min(sample_info.priority), dtype=tf.int64)

      if self._allow_variable_length_episodes:
        # Filter off policy samples.
        return tf.math.equal(self._model_id, data_model_id)
      else:
        # Filter infeasible placements with shorter episode lengths than
        # expected along with off policy samples.
        data = sample.data
        return tf.math.logical_and(
            tf.math.equal(tf.size(data.discount), self._sequence_length),
            tf.math.equal(self._model_id, data_model_id))

    def make_dataset(_):
      # `experience_dataset_fn` returns a tf.Dataset. Each item is a (Trajectory
      # , SampleInfo) tuple, and the Trajectory represents one single episode
      # of a fixed sequence length. The Trajectory dimensions are [1, T, ...].
      train_dataset = self._experience_dataset_fn()
      train_dataset = train_dataset.filter(_filter_invalid_episodes)
      if self._per_sequence_fn:
        train_dataset = train_dataset.map(
            self._per_sequence_fn,
            num_parallel_calls=tf.data.AUTOTUNE,
            deterministic=False)

      # We unbatch the dataset shaped [B, T, ...] to a new dataset that
      # contains individual elements.
      # Note that we unbatch across the time dimension, which could result
      # in mini batches that contain subsets from more than one sequences.
      # PPO agent can handle mini batches across episode boundaries.
      train_dataset = train_dataset.unbatch()
      train_dataset = train_dataset.batch(1, drop_remainder=True)
      train_dataset = train_dataset.shuffle(self._shuffle_buffer_size)
      #train_dataset = train_dataset.take(self._num_samples)
      #train_dataset = train_dataset.cache()
      train_dataset = train_dataset.repeat(self._num_epochs)
      train_dataset = train_dataset.batch(
          self._minibatch_size, drop_remainder=True)
      #train_dataset = train_dataset.repeat()

      options = tf.data.Options()
      options.deterministic = False
      options.experimental_optimization.parallel_batch = True
      train_dataset = train_dataset.with_options(options)

      return train_dataset

    with self._strategy.scope():

      if self._strategy.num_replicas_in_sync > 1:
        self._train_dataset = (
            self._strategy.distribute_datasets_from_function(make_dataset))
      else:
        self._train_dataset = make_dataset(0)
      self._train_iterator = iter(self._train_dataset)

  def _get_train_steps_per_iteration(self):
    """Number of train steps each time learner.run() is called."""

    # We exhaust all num_episodes_per_iteration taken from Reverb in this setup.
    # Here we assume that there's only 1 episode per batch, and each episode is
    # of the fixed sequence length.
    num_mini_batches = int(self._num_samples * self._num_epochs /
                           self._minibatch_size)
    train_steps = int(num_mini_batches / self.num_replicas)
    return train_steps

  def wait_for_data(self):
    """Blocking call on dataset."""
    traj, sample_info = next(self._train_iterator)
    del traj
    del sample_info

  def run(self):
    """Train `num_episodes_per_iteration` repeating for `num_epochs`.

    Returns:
      The total loss computed before running the final step.
    """
    loss_info = self._generic_learner.run(self._steps_per_iter,
                                          self._train_iterator)
    self._model_id.assign_add(1)
    self._create_datasets()
    return loss_info

  @property
  def train_step_numpy(self):
    """The current train_step.

    Returns:
      The current `train_step`. Note this will return a scalar numpy array which
      holds the `train_step` value when this was called.
    """
    return self._generic_learner.train_step_numpy



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Library for combining training corpora."""

import os
import json

from absl import logging

import tensorflow as tf

_FILE_NAME = 'corpus_description.json'


def combine_corpus(root_dir: str) -> None:
  module_names = []
  output_corpus_description = {}

  for sub_dir in tf.io.gfile.listdir(root_dir):
    path = os.path.join(root_dir, sub_dir, _FILE_NAME)

    logging.info('processing %s', path)

    if not tf.io.gfile.exists(path):
      logging.error('%s does not exist.', path)
      continue

    with tf.io.gfile.GFile(path, 'r') as f:
      corpus_description = json.load(f)
      module_names.extend([
          os.path.join(sub_dir, name) for name in corpus_description['modules']
      ])
      del corpus_description['modules']
      if len(output_corpus_description) == 0:
        output_corpus_description = corpus_description
      elif corpus_description != output_corpus_description:
        raise ValueError('Input corpora differ by more than modules.')

  output_corpus_description['modules'] = module_names

  with tf.io.gfile.GFile(os.path.join(root_dir, _FILE_NAME), 'w') as f:
    json.dump(output_corpus_description, f, indent=2)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module for collect data of inlining-for-size."""

import os
import tempfile
from typing import Dict, Tuple

import gin
import tensorflow as tf

from compiler_opt.rl import compilation_runner
from compiler_opt.rl import corpus
from compiler_opt.rl import log_reader

_DEFAULT_IDENTIFIER = 'default'


@gin.configurable(module='runners')
class InliningRunner(compilation_runner.CompilationRunner):
  """Class for collecting data for inlining-for-size.

  Usage:
  inliner = InliningRunner(
                clang_path, llvm_size_path, launcher_path,
                moving_average_decay_rate)
  serialized_sequence_example, default_reward, moving_average_reward,
  policy_reward = inliner.collect_data(
      ir_path, tf_policy_path, default_reward, moving_average_reward)
  """

  def __init__(self, llvm_size_path: str, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self._llvm_size_path = llvm_size_path

  def compile_fn(
      self, command_line: corpus.FullyQualifiedCmdLine, tf_policy_path: str,
      reward_only: bool,
      workdir: str) -> Dict[str, Tuple[tf.train.SequenceExample, float]]:
    """Run inlining for the given IR file under the given policy.

    Args:
      command_line: the fully qualified command line.
      tf_policy_path: path to TF policy direcoty on local disk.
      reward_only: whether only return native size.

    Returns:
      A dict mapping from example identifier to tuple containing:
        sequence_example: A tf.SequenceExample proto describing compilation
        trace, None if reward_only == True.
        native_size: Native size of the final native code.

    Raises:
      subprocess.CalledProcessError: if process fails.
      compilation_runner.ProcessKilledError: (which it must pass through) on
      cancelled work.
      RuntimeError: if llvm-size produces unexpected output.
    """

    working_dir = tempfile.mkdtemp(dir=workdir)

    log_path = os.path.join(working_dir, 'log')
    output_native_path = os.path.join(working_dir, 'native')

    native_size = 0
    cmdline = []
    if self._launcher_path:
      cmdline.append(self._launcher_path)
    cmdline.extend([self._clang_path] + list(command_line) + [
        '-mllvm', '-enable-ml-inliner=development', '-mllvm', '-training-log=' +
        log_path, '-o', output_native_path
    ])
    if tf_policy_path:
      cmdline.extend(
          ['-mllvm', '-ml-inliner-model-under-training=' + tf_policy_path])
    compilation_runner.start_cancellable_process(cmdline,
                                                 self._compilation_timeout,
                                                 self._cancellation_manager)
    cmdline = [self._llvm_size_path, output_native_path]
    output_bytes = compilation_runner.start_cancellable_process(
        cmdline,
        timeout=self._compilation_timeout,
        cancellation_manager=self._cancellation_manager,
        want_output=True)
    if not output_bytes:
      raise RuntimeError(f'Empty llvm-size output: {" ".join(cmdline)}')
    output = output_bytes.decode('utf-8')
    tmp = output.split('\n')
    if len(tmp) != 3:
      raise RuntimeError(f'Wrong llvm-size output {output}')
    tmp = tmp[1].split('\t')
    native_size = int(tmp[0])

    if native_size == 0:
      return {}

    if reward_only:
      return {_DEFAULT_IDENTIFIER: (None, native_size)}

    result = log_reader.read_log_as_sequence_examples(log_path)
    if len(result) != 1:
      return {}
    sequence_example = next(iter(result.values()))

    if not sequence_example.HasField('feature_lists'):
      return {}

    return {_DEFAULT_IDENTIFIER: (sequence_example, native_size)}



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for compiler_opt.tools.make_corpus_lib"""

import json
import os

from absl.testing import absltest

from compiler_opt.tools import make_corpus_lib


class MakeCorpusTest(absltest.TestCase):

  def test_load_bitcode_from_directory(self):
    outer = self.create_tempdir()
    tempdir = outer.mkdir(dir_path='nested')
    tempdir.create_file('test1.bc')
    tempdir.create_file('test2.bc')
    relative_paths = make_corpus_lib.load_bitcode_from_directory(outer)
    relative_paths = sorted(relative_paths)
    self.assertEqual(relative_paths[0], 'nested/test1.bc')
    self.assertEqual(relative_paths[1], 'nested/test2.bc')

  def test_copy_bitcode(self):
    build_dir = self.create_tempdir()
    nested_dir = build_dir.mkdir(dir_path='nested')
    nested_dir.create_file('test1.bc')
    nested_dir.create_file('test2.bc')
    relative_paths = ['nested/test1.bc', 'nested/test2.bc']
    corpus_dir = self.create_tempdir()
    make_corpus_lib.copy_bitcode(relative_paths, build_dir, corpus_dir)
    output_files = sorted(os.listdir(os.path.join(corpus_dir, './nested')))
    self.assertEqual(output_files[0], 'test1.bc')
    self.assertEqual(output_files[1], 'test2.bc')

  def test_write_corpus_manifest(self):
    relative_output_paths = ['test/test1.bc', 'test/test2.bc']
    output_dir = self.create_tempdir()
    default_args = ['-O3', '-c']
    make_corpus_lib.write_corpus_manifest(relative_output_paths, output_dir,
                                          default_args)
    with open(
        os.path.join(output_dir, 'corpus_description.json'),
        encoding='utf-8') as corpus_description_file:
      corpus_description = json.load(corpus_description_file)
    self.assertEqual(corpus_description['global_command_override'],
                     default_args)
    self.assertEqual(corpus_description['has_thinlto'], False)
    self.assertEqual(corpus_description['modules'], relative_output_paths)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test for buffered_scheduler."""

import concurrent.futures
import threading
import time

from absl.testing import absltest
from compiler_opt.distributed import worker
from compiler_opt.distributed import buffered_scheduler
from compiler_opt.distributed.local import local_worker_manager


class BufferedSchedulerTest(absltest.TestCase):

  def test_canonical_futures(self):

    class WaitingWorker(worker.Worker):

      def wait_seconds(self, n: int):
        time.sleep(n)
        return n + 1

    with local_worker_manager.LocalWorkerPoolManager(WaitingWorker, 2) as pool:
      _, futures = buffered_scheduler.schedule_on_worker_pool(
          lambda w, v: w.wait_seconds(v), range(4), pool)
      not_done = futures
      entered_count = 0
      while not_done:
        _, not_done = concurrent.futures.wait(not_done, timeout=0.5)
        entered_count += 1
      self.assertGreater(entered_count, 1)
      self.assertListEqual([r.result() for r in futures], list(range(1, 5)))

  def test_simple_scheduling(self):

    class TheWorker(worker.Worker):

      def square(self, the_value, extra_factor=1):
        return the_value * the_value * extra_factor

    with local_worker_manager.LocalWorkerPoolManager(TheWorker, 2) as pool:
      workers, futures = buffered_scheduler.schedule_on_worker_pool(
          lambda w, v: w.square(v), range(10), pool)
      self.assertLen(workers, 2)
      self.assertLen(futures, 10)
      worker.wait_for(futures)
      self.assertListEqual([f.result() for f in futures],
                           [x * x for x in range(10)])

      _, futures = buffered_scheduler.schedule_on_worker_pool(
          lambda w, v: w.square(**v), [dict(the_value=v) for v in range(10)],
          pool)
      worker.wait_for(futures)
      self.assertListEqual([f.result() for f in futures],
                           [x * x for x in range(10)])

      # same idea, but mix some kwargs
      _, futures = buffered_scheduler.schedule_on_worker_pool(
          lambda w, v: w.square(v[0], **v[1]),
          [(v, dict(extra_factor=10)) for v in range(10)], pool)
      worker.wait_for(futures)
      self.assertListEqual([f.result() for f in futures],
                           [x * x * 10 for x in range(10)])

  def test_schedules(self):
    call_count = [0] * 4
    locks = [threading.Lock() for _ in range(4)]

    def wkr_factory(i):

      def wkr():
        with locks[i]:
          call_count[i] += 1

      return wkr

    wkrs = [wkr_factory(i) for i in range(4)]

    def job(wkr):
      future = concurrent.futures.Future()

      def task():
        wkr()
        future.set_result(0)

      threading.Timer(interval=0.10, function=task).start()
      return future

    work = [job] * 20

    worker.wait_for(buffered_scheduler.schedule(work, wkrs))
    self.assertEqual(sum(call_count), 20)

  def test_balances(self):
    call_count = [0] * 4
    locks = [threading.Lock() for _ in range(4)]

    def wkr_factory(i):

      def wkr():
        with locks[i]:
          call_count[i] += 1

      return wkr

    def slow_wkr():
      with locks[0]:
        call_count[0] += 1
      time.sleep(1)

    wkrs = [slow_wkr] + [wkr_factory(i) for i in range(1, 4)]

    def job(wkr):
      future = concurrent.futures.Future()

      def task():
        wkr()
        future.set_result(0)

      threading.Timer(interval=0.10, function=task).start()
      return future

    work = [job] * 20

    worker.wait_for(buffered_scheduler.schedule(work, wkrs, buffer=2))
    self.assertEqual(sum(call_count), 20)
    # since buffer=2, 2 tasks get assigned to the slow wkr, the rest
    # should've been assigned elsewhere if load balancing works.
    self.assertEqual(call_count[0], 2)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Local Process Pool - based middleware implementation.

This is a simple implementation of a worker pool, running on the local machine.
Each worker object is hosted by a separate process. Each worker object may
handle a number of concurrent requests. The client is given a stub object that
exposes the same methods as the worker, just that they return Futures.

There is a bidirectional pipe between a stub and its corresponding
process/worker. One direction is used to place tasks (method calls), the other
to place results. Tasks and results are correlated by a monotonically
incrementing counter maintained by the stub.

The worker process dequeues tasks promptly and either re-enqueues them to a
local thread pool, or, if the task is 'urgent', it executes it promptly.
"""
import concurrent.futures
import dataclasses
import functools
import multiprocessing
import psutil
import threading

from absl import logging
# pylint: disable=unused-import
from compiler_opt.distributed import worker

from contextlib import AbstractContextManager
from multiprocessing import connection
from typing import Any, Callable, Dict, List, Optional


@dataclasses.dataclass(frozen=True)
class Task:
  msgid: int
  func_name: str
  args: tuple
  kwargs: dict
  is_urgent: bool


@dataclasses.dataclass(frozen=True)
class TaskResult:
  msgid: int
  success: bool
  value: Any


def _run_impl(pipe: connection.Connection, worker_class: 'type[worker.Worker]',
              *args, **kwargs):
  """Worker process entrypoint."""

  # A setting of 1 does not inhibit the while loop below from running since
  # that runs on the main thread of the process. Urgent tasks will still
  # process near-immediately. `threads` only controls how many threads are
  # spawned at a time which execute given tasks. In the typical clang-spawning
  # jobs, this effectively limits the number of clang instances spawned.
  pool = concurrent.futures.ThreadPoolExecutor(max_workers=1)
  obj = worker_class(*args, **kwargs)

  # Pipes are not thread safe
  pipe_lock = threading.Lock()

  def send(task_result: TaskResult):
    with pipe_lock:
      pipe.send(task_result)

  def make_ondone(msgid):

    def on_done(f: concurrent.futures.Future):
      if f.exception():
        send(TaskResult(msgid=msgid, success=False, value=f.exception()))
      else:
        send(TaskResult(msgid=msgid, success=True, value=f.result()))

    return on_done

  # Run forever. The stub will just kill the runner when done.
  while True:
    task: Task = pipe.recv()
    the_func = getattr(obj, task.func_name)
    application = functools.partial(the_func, *task.args, **task.kwargs)
    if task.is_urgent:
      try:
        res = application()
        send(TaskResult(msgid=task.msgid, success=True, value=res))
      except BaseException as e:  # pylint: disable=broad-except
        send(TaskResult(msgid=task.msgid, success=False, value=e))
    else:
      pool.submit(application).add_done_callback(make_ondone(task.msgid))


def _run(pipe: connection.Connection, worker_class: 'type[worker.Worker]',
         *args, **kwargs):
  try:
    _run_impl(pipe, worker_class, *args, **kwargs)
  except BaseException as e:
    logging.error(e)
    raise e


def _make_stub(cls: 'type[worker.Worker]', *args, **kwargs):

  class _Stub:
    """Client stub to a worker hosted by a process."""

    def __init__(self):
      parent_pipe, child_pipe = multiprocessing.get_context().Pipe()
      self._pipe = parent_pipe
      self._pipe_lock = threading.Lock()

      # this is the process hosting one worker instance.
      # we set aside 1 thread to coordinate running jobs, and the main thread
      # to handle high priority requests. The expectation is that the user
      # achieves concurrency through multiprocessing, not multithreading.
      self._process = multiprocessing.get_context().Process(
          target=functools.partial(_run, child_pipe, cls, *args, **kwargs))
      # lock for the msgid -> reply future map. The map will be set to None
      # when we stop.
      self._lock = threading.Lock()
      self._map: Dict[int, concurrent.futures.Future] = {}

      # thread draining the pipe
      self._pump = threading.Thread(target=self._msg_pump)

      # Set the state of this worker to "dead" if the process dies naturally.
      def observer():
        self._process.join()
        # Feed the parent pipe a poison pill, this kills msg_pump
        child_pipe.send(None)

      self._observer = threading.Thread(target=observer)

      # atomic control to _msgid
      self._msgidlock = threading.Lock()
      self._msgid = 0

      # start the worker and the message pump
      self._process.start()
      # the observer must follow the process start, otherwise join() raises.
      self._observer.start()
      self._pump.start()

    def _msg_pump(self):
      while True:
        task_result: Optional[TaskResult] = self._pipe.recv()
        if task_result is None:  # Poison pill fed by observer
          break
        with self._lock:
          future = self._map[task_result.msgid]
          del self._map[task_result.msgid]
        # The following will trigger any callbacks defined on the future, as a
        # direct function call. If those callbacks were set by the scheduler,
        # it's important that self._lock isn't being held when they are being
        # called, otherwise a deadlock could arise from __get_attr__ trying to
        # acquire the lock.
        if task_result.success:
          future.set_result(task_result.value)
        else:
          future.set_exception(task_result.value)

      # clear out pending futures and mark ourselves as "stopped" by null-ing
      # the map
      with self._lock:
        for _, v in self._map.items():
          v.set_exception(concurrent.futures.CancelledError())
        self._map = None

    def _is_stopped(self):
      return self._map is None

    def __getattr__(self, name) -> Callable[[Any], concurrent.futures.Future]:
      result_future = concurrent.futures.Future()

      with self._msgidlock:
        msgid = self._msgid
        self._msgid += 1

      def remote_call(*args, **kwargs):
        with self._lock:
          if self._is_stopped():
            result_future.set_exception(concurrent.futures.CancelledError())
          else:
            with self._pipe_lock:
              self._pipe.send(
                  Task(
                      msgid=msgid,
                      func_name=name,
                      args=args,
                      kwargs=kwargs,
                      is_urgent=cls.is_priority_method(name)))
            self._map[msgid] = result_future
        return result_future

      return remote_call

    def shutdown(self):
      try:
        # Killing the process triggers observer exit, which triggers msg_pump
        # exit
        self._process.kill()
      except:  # pylint: disable=bare-except
        pass

    def set_nice(self, val: int):
      """Sets the nice-ness of the process, this modifies how the OS
      schedules it. Only works on Unix, since val is presumed to be an int.
      """
      psutil.Process(self._process.pid).nice(val)

    def set_affinity(self, val: List[int]):
      """Sets the CPU affinity of the process, this modifies which cores the OS
      schedules it on.
      """
      psutil.Process(self._process.pid).cpu_affinity(val)

    def join(self):
      self._observer.join()
      self._pump.join()
      self._process.join()

    def __dir__(self):
      return [n for n in dir(cls) if not n.startswith('_')]

  return _Stub()


def create_local_worker_pool(worker_cls: 'type[worker.Worker]',
                             count: Optional[int], *args,
                             **kwargs) -> worker.FixedWorkerPool:
  """Create a local worker pool for worker_cls."""
  if not count:
    count = multiprocessing.get_context().cpu_count()
  final_kwargs = worker.get_full_worker_args(worker_cls, **kwargs)
  stubs = [_make_stub(worker_cls, *args, **final_kwargs) for _ in range(count)]
  return worker.FixedWorkerPool(workers=stubs, worker_concurrency=16)


def close_local_worker_pool(pool: worker.FixedWorkerPool):
  """Close the given LocalWorkerPool."""
  # first, trigger killing the worker process and exiting of the msg pump,
  # which will also clear out any pending futures.
  for stub in pool.get_currently_active():
    stub.shutdown()
  # now wait for the message pumps to indicate they exit.
  for stub in pool.get_currently_active():
    stub.join()


class LocalWorkerPoolManager(AbstractContextManager):
  """A pool of workers hosted on the local machines, each in its own process."""

  def __init__(self, worker_class: 'type[worker.Worker]', count: Optional[int],
               *args, **kwargs):
    self._pool = create_local_worker_pool(worker_class, count, *args, **kwargs)

  def __enter__(self) -> worker.FixedWorkerPool:
    return self._pool

  def __exit__(self, *args):
    close_local_worker_pool(self._pool)

  def __del__(self):
    self.__exit__()

  @property
  def stubs(self):
    # Return a shallow copy, to avoid something messing the internal list up
    return list(self._stubs)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the feature_importance_utils.py module"""

from absl.testing import absltest

from compiler_opt.tools import feature_importance_utils

import tensorflow as tf
import numpy

from types import SimpleNamespace

test_tensor_spec = {
    'test1': tf.constant([1, 2], dtype=tf.int64),
    'test2': tf.constant([3, 4], dtype=tf.int64)
}


class FeatureImportanceTest(absltest.TestCase):

  def test_get_input_signature(self):
    input_signature = feature_importance_utils.get_input_signature(
        test_tensor_spec)
    self.assertListEqual(list(input_signature.keys()), ['test1', 'test2'])
    self.assertListEqual(input_signature['test1'][0].tolist(), [2])
    self.assertEqual(input_signature['test1'][1], tf.int64)

  def test_get_signature_total_size(self):
    input_signature = feature_importance_utils.get_input_signature(
        test_tensor_spec)
    total_size = feature_importance_utils.get_signature_total_size(
        input_signature)
    self.assertEqual(total_size, 4)

  def test_packing_flattening(self):
    input_signature = feature_importance_utils.get_input_signature(
        test_tensor_spec)
    total_size = feature_importance_utils.get_signature_total_size(
        input_signature)
    flattened_input = feature_importance_utils.flatten_input(
        test_tensor_spec, total_size)
    packed_input = feature_importance_utils.pack_flat_array_into_input(
        flattened_input, input_signature)
    self.assertListEqual(list(packed_input.keys()), ['test1', 'test2'])
    self.assertTrue(
        numpy.array_equal(test_tensor_spec['test1'].numpy(),
                          packed_input['test1'].numpy()))
    self.assertTrue(
        numpy.array_equal(test_tensor_spec['test2'].numpy(),
                          packed_input['test2'].numpy()))

  def test_trajectory_processing(self):
    batched_test_tensor_spec = {
        'test1': tf.constant([[1, 2]]),
        'test2': tf.constant([[3, 4]])
    }
    pre_trajectory = {
        'observation': batched_test_tensor_spec,
        'step_type': tf.constant([[1]]),
        'reward': tf.constant([[2]]),
        'discount': tf.constant([[3]])
    }
    trajectory = SimpleNamespace(**pre_trajectory)
    processed_trajectory = feature_importance_utils.process_raw_trajectory(
        trajectory)
    self.assertListEqual(
        list(processed_trajectory.keys()),
        ['test1', 'test2', 'step_type', 'reward', 'discount'])
    # make sure values are correct
    self.assertTrue(
        numpy.array_equal(processed_trajectory['test1'].numpy(), [1, 2]))
    self.assertTrue(
        numpy.array_equal(processed_trajectory['test2'].numpy(), [3, 4]))
    # make sure the tensors got squeezed
    self.assertTrue(
        numpy.array_equal(tf.shape(processed_trajectory['test1']).numpy(), [2]))
    self.assertTrue(
        numpy.array_equal(
            tf.shape(processed_trajectory['discount']).numpy(), [1]))

  def test_get_max_part_size(self):
    ragged_nested_tensor_spec = {
        'test1': tf.constant([1, 2, 3, 4]),
        'test2': tf.constant([1, 2])
    }
    input_signature = feature_importance_utils.get_input_signature(
        ragged_nested_tensor_spec)
    max_part_size = feature_importance_utils.get_max_part_size(input_signature)
    self.assertEqual(max_part_size, 4)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Merge multiple best trajectory repo into one."""

import json

from absl import app
from absl import flags
from absl import logging

from compiler_opt.rl import best_trajectory

_BEST_TRAJECTORY_PATHS = flags.DEFINE_multi_string(
    'best_trajectory_paths', '',
    'best trajectory repo dump to be merged in json format.')
_OUTPUT_JSON_PATH = flags.DEFINE_string(
    'output_json_path', '',
    'output path of the merged best trajectory repo in json format if given.')
_OUTPUT_CSV_PATH = flags.DEFINE_string(
    'output_csv_path', '',
    'output path of the merged best trajectory repo in csv format if given.')

FLAGS = flags.FLAGS


def main(argv):
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  # we don't use action_name in the merging process here, so just set it to
  # empty string.
  merged_best_trajectory_repo = best_trajectory.BestTrajectoryRepo(
      action_name='')

  for path in _BEST_TRAJECTORY_PATHS.value:
    logging.info('merging repo: %s', path)
    tmp = best_trajectory.BestTrajectoryRepo(action_name='')
    # The json file is broken sometimes.
    # Open issue: https://github.com/google/ml-compiler-opt/issues/163
    try:
      tmp.load_from_json_file(path)
      merged_best_trajectory_repo.combine_with_other_repo(tmp)
    except json.decoder.JSONDecodeError:
      logging.error('failed to load input repo: %s', path)

  if _OUTPUT_JSON_PATH.value:
    tmp.sink_to_json_file(_OUTPUT_JSON_PATH.value)

  if _OUTPUT_CSV_PATH.value:
    tmp.sink_to_csv_file(_OUTPUT_CSV_PATH.value)


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generate test model given a problem and an algorithm."""

import os

from absl import app
from absl import flags
from absl import logging

import gin

from compiler_opt.rl import agent_config
from compiler_opt.rl import gin_external_configurables  # pylint: disable=unused-import
from compiler_opt.rl import policy_saver
from compiler_opt.rl import registry

flags.DEFINE_string('root_dir', os.getenv('TEST_UNDECLARED_OUTPUTS_DIR'),
                    'Root directory for writing saved models.')
flags.DEFINE_multi_string('gin_files', [],
                          'List of paths to gin configuration files.')
flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files.')

FLAGS = flags.FLAGS


@gin.configurable
def generate_test_model(agent_config_type=agent_config.PPOAgentConfig):
  """Generate test model."""
  root_dir = FLAGS.root_dir

  problem_config = registry.get_configuration()
  time_step_spec, action_spec = problem_config.get_signature_spec()
  preprocessing_layer_creator = problem_config.get_preprocessing_layer_creator()

  # Initialize trainer and policy saver.
  tf_agent = agent_config.create_agent(
      agent_config_type(time_step_spec=time_step_spec, action_spec=action_spec),
      preprocessing_layer_creator=preprocessing_layer_creator)

  policy_dict = {
      'saved_policy': tf_agent.policy,
      'saved_collect_policy': tf_agent.collect_policy,
  }
  saver = policy_saver.PolicySaver(policy_dict=policy_dict)

  # Save policy.
  saver.save(root_dir)


def main(_):
  gin.parse_config_files_and_bindings(
      FLAGS.gin_files, bindings=FLAGS.gin_bindings, skip_unknown=True)
  logging.info(gin.config_str())

  generate_test_model()


if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LLVM Policy Trainer."""

import time

from absl import logging

import gin
import tensorflow as tf
from compiler_opt.rl import random_net_distillation
from tf_agents.agents import tf_agent
from tf_agents.policies import policy_loader

from tf_agents.utils import common as common_utils
from typing import Optional


@gin.configurable
class Trainer(object):
  """Object that trains LLVM policy.

  After initialization, the function 'train' can be called multiple times to
  train on different datasets. An example usage:

  ```python
  trainer = Trainer(root_dir, agent)
  trainer.train(data_iter_1, num_iterations_1)
  trainer.train(data_iter_2, num_iterations_2)
  ```
  """

  def __init__(
      self,
      root_dir: str,
      agent: tf_agent.TFAgent,
      random_network_distillation: Optional[
          random_net_distillation.RandomNetworkDistillation] = None,
      warmstart_policy_dir: Optional[str] = None,
      # Params for summaries and logging
      checkpoint_interval=10000,
      log_interval=100,
      summary_log_interval=100,
      summary_export_interval=1000,
      summaries_flush_secs=10):
    """Initialize the Trainer object.

    Args:
      root_dir: str, the root directory to host all required sub-directories.
      agent: a tf_agents.agents.TFAgent object.
      random_network_distillation: a random_net_distillation.RND object.
      warmstart_policy_dir: the directory to warmstart the policy if given.
      checkpoint_interval: int, the training step interval for saving
        checkpoint.
      log_interval: int, the training step interval for logging.
      summary_log_interval: the number of steps in between logging metrics
        to tensorboard.
      summary_export_interval: int, the training step interval for exporting
        to tensorboard.
      summaries_flush_secs: int, the seconds for flushing to tensorboard.
    """
    self._root_dir = root_dir
    self._agent = agent
    self._random_network_distillation = random_network_distillation
    self._checkpoint_interval = checkpoint_interval
    self._log_interval = log_interval
    self._summary_log_interval = summary_log_interval
    self._summary_export_interval = summary_export_interval

    self._summary_writer = tf.summary.create_file_writer(
        self._root_dir, flush_millis=summaries_flush_secs * 1000)
    self._summary_writer.set_as_default()

    self._global_step = tf.compat.v1.train.get_or_create_global_step()

    # Initialize agent and trajectory replay.
    # Wrap training and trajectory replay in a tf.function to make it much
    # faster.
    self._agent.initialize()
    self._agent.train = common_utils.function(self._agent.train)
    if self._random_network_distillation:
      self._random_network_distillation.train = common_utils.function(
          self._random_network_distillation.train)

    self._initialize_metrics()

    # Load warmstart policy before restoring from checkpoint.
    if warmstart_policy_dir:
      warmstart_policy = policy_loader.load(warmstart_policy_dir)
      self._agent.policy.update(
          policy=warmstart_policy,
          tau=1.0,
          tau_non_trainable=None,
          sort_variables_by_name=False)

    self._checkpointer = common_utils.Checkpointer(
        ckpt_dir=self._root_dir,
        agent=self._agent,
        global_step=self._global_step)
    self._checkpointer.initialize_or_restore()

    self._start_time = time.time()

  def _initialize_metrics(self):
    """Initializes metrics."""
    self._data_action_mean = tf.keras.metrics.Mean()
    self._data_reward_mean = tf.keras.metrics.Mean()
    self._num_trajectories = tf.keras.metrics.Sum()

  def _update_metrics(self, experience, monitor_dict):
    """Updates metrics and exports to Tensorboard."""
    if tf.math.equal(self._global_step % self._summary_log_interval, 0):
      is_action = ~experience.is_boundary()

      self._data_action_mean.update_state(
          experience.action, sample_weight=is_action)
      self._data_reward_mean.update_state(
          experience.reward, sample_weight=is_action)
      self._num_trajectories.update_state(experience.is_first())

    # Check earlier rather than later if we should record summaries.
    # TF also checks it, but much later. Needed to avoid looping through
    # the dict so gave the if a bigger scope
    if tf.summary.should_record_summaries():
      with tf.name_scope('default/'):
        tf.summary.scalar(
            name='data_action_mean',
            data=self._data_action_mean.result(),
            step=self._global_step)
        tf.summary.scalar(
            name='data_reward_mean',
            data=self._data_reward_mean.result(),
            step=self._global_step)
        tf.summary.scalar(
            name='num_trajectories',
            data=self._num_trajectories.result(),
            step=self._global_step)

      for name_scope, d in monitor_dict.items():
        with tf.name_scope(name_scope + '/'):
          for key, value in d.items():
            tf.summary.scalar(name=key, data=value, step=self._global_step)

      tf.summary.histogram(
          name='reward', data=experience.reward, step=self._global_step)

  def _reset_metrics(self):
    """Reset num_trajectories."""
    self._num_trajectories.reset_states()

  def _log_experiment(self, loss):
    """Log training info."""
    if tf.math.equal(self._global_step % self._log_interval, 0):
      global_step_val = self._global_step.numpy()
      logging.info('step = %d, loss = %g', global_step_val, loss)
      time_acc = time.time() - self._start_time
      steps_per_sec = self._log_interval / time_acc
      logging.info('%.3f steps/sec', steps_per_sec)
      self._start_time = time.time()

  def _save_checkpoint(self):
    if tf.math.equal(self._global_step % self._checkpoint_interval, 0):
      self._checkpointer.save(global_step=self._global_step)

  def global_step_numpy(self):
    return self._global_step.numpy()

  def train(self, dataset_iter, monitor_dict, num_iterations: int):
    """Trains policy with data from dataset_iter for num_iterations steps."""
    self._reset_metrics()
    # context management is implemented in decorator
    # pylint: disable=not-context-manager
    with tf.summary.record_if(lambda: tf.math.equal(
        self._global_step % self._summary_export_interval, 0)):
      for _ in range(num_iterations):
        # When the data is not enough to fill in a batch, next(dataset_iter)
        # will throw StopIteration exception, logging a warning message instead
        # of killing the training when it happens.
        try:
          experience = next(dataset_iter)
        except StopIteration:
          logging.warning(
              ('Warning: skip training because do not have enough data to fill '
               'in a batch, consider increase data or reduce batch size.'))
          break

        # random network distillation for intrinsic reward generation
        if self._random_network_distillation:
          experience = self._random_network_distillation.train(experience)

        loss = self._agent.train(experience)

        self._update_metrics(experience, monitor_dict)
        self._log_experiment(loss.loss)
        self._save_checkpoint()



# coding=utf-8
# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for the feature_importance.py script

Refactored into a separate script so that we can run all of these utilities
through pytest without needing to add any odd conditionals to deal with
duplicate absl flags etc.
"""

import tensorflow as tf
import numpy
import numpy.typing

from tf_agents.typing import types
from typing import Callable, Dict, Tuple

SignatureType = Dict[str, Tuple[numpy.typing.ArrayLike, tf.dtypes.DType]]


def get_input_signature(example_input: types.NestedTensorSpec) -> SignatureType:
  """Gets the signature of an observation

  This function takes in an example input and returns a signature of that
  input containing all of the info needed to restructure a flat array back into
  the original format later on. This function returns a dictionary with the
  same keys as the original input but with the items being tuples where the
  first value is the shape of that feature and the second is its data type.

  Args:
    example_input: a nested tensor spec (dictionary of tensors) that serves
      as an example for generating the signature.
  """
  input_signature = {}
  for input_key in example_input:
    input_signature[input_key] = (tf.shape(example_input[input_key]).numpy(),
                                  example_input[input_key].dtype)
  return input_signature


def get_signature_total_size(input_signature: SignatureType) -> int:
  """Gets the total number of elements in a single problem instance

  Args:
    input_signature: An input signature to calculate the number of elements in
  """
  total_size = 0
  for input_key in input_signature:
    total_size += numpy.prod(input_signature[input_key][0])
  return total_size


def pack_flat_array_into_input(
    flat_array: numpy.typing.ArrayLike,
    signature_spec: SignatureType) -> types.NestedTensorSpec:
  """Packs a flat array into a nested tensor spec to feed into a model

  Args:
    flat_array: The data to be packed back into the specified nested tensor
      specification
    signature_spec: A signature that is used to create the correct structure
      for all of the values in the flat array
  """
  output_input_dict = {}
  current_index = 0
  for needed_input in signature_spec:
    part_size = numpy.prod(signature_spec[needed_input][0])
    needed_subset = flat_array[current_index:current_index + part_size]
    current_index += part_size
    output_input_dict[needed_input] = tf.cast(
        tf.constant(needed_subset, shape=signature_spec[needed_input][0]),
        dtype=signature_spec[needed_input][1])
  return output_input_dict


def flatten_input(to_flatten: types.NestedTensorSpec,
                  array_size: int) -> numpy.typing.ArrayLike:
  """Flattens problem instance data into a flat array for shap

  Args:
    to_flatten: A nested tensor spec of data that needs to be flattend into
      an array
    array_size: An integer representing the size of the output array. Used for
      allocating the flat array to place all the data in.
  """
  output_array = numpy.empty(array_size)
  input_index = 0
  for input_key in to_flatten:
    current_size = tf.size(to_flatten[input_key])
    end_index = input_index + current_size
    output_array[input_index:end_index] = to_flatten[input_key].numpy().astype(
        numpy.float32)
    input_index += current_size
  return output_array


def process_raw_trajectory(
    raw_trajectory: types.ForwardRef) -> types.NestedTensorSpec:
  """Processes the raw example data into a nested tensor spec that can be
  easily fed into a model.

  Args:
    raw_trajectory: Raw data representing an individual problem instance from
      a trace.
  """
  observation = raw_trajectory.observation
  observation.update({
      'step_type': raw_trajectory.step_type,
      'reward': raw_trajectory.reward,
      'discount': raw_trajectory.discount
  })

  # remove batch size dimension
  for key in observation:
    observation[key] = tf.squeeze(observation[key], axis=0)

  return observation


def collapse_values(input_signature: SignatureType,
                    shap_values: numpy.typing.ArrayLike,
                    num_examples: int) -> numpy.typing.ArrayLike:
  """Collapses shap values so that there is only a single value per feature

  Args:
    input_signature: The signature of the model input. Used to determine what
      (if any) features need to be collapsed.
    shap_values: A numpy array of shap values that need to be processed.
  """
  output_shap_values = numpy.empty((num_examples, len(input_signature)))
  for i in range(0, num_examples):
    current_index = 0
    current_feature = 0
    for input_key in input_signature:
      part_size = numpy.prod(input_signature[input_key][0])
      output_shap_values[i, current_feature] = numpy.sum(
          shap_values[i, current_index:current_index + part_size])
      current_feature += 1
      current_index += part_size
  return output_shap_values


def get_max_part_size(input_signature: SignatureType) -> int:
  """Gets the size (as a single scalar) of the largest feature in terms of
  the number of elements.

  Args:
    input_signature: The input signature that we want to find the largest
      feature in.
  """
  part_sizes = numpy.empty(len(input_signature))
  for index, input_key in enumerate(input_signature):
    part_sizes[index] = numpy.prod(input_signature[input_key][0])
  return numpy.max(part_sizes)


def create_run_model_function(action_fn: Callable,
                              input_sig: SignatureType) -> Callable:
  """Returns a function that takes in a flattend input array and returns the
  model output as a scalar.

  Args:
    action_fn: The action function from the tensorflow saved model saved
      through tf_agents
    input_sig: The input signature for the model currently under analysis.
      Used to pack the flat array back into a nested tensor spec.
  """

  def run_model(flat_input_array):
    output = numpy.empty(flat_input_array.shape[0])
    for index, flat_input in enumerate(flat_input_array):
      input_dict = pack_flat_array_into_input(flat_input, input_sig)
      model_output = action_fn(**input_dict).items()
      # get the value of the first item as a numpy array
      output[index] = list(model_output)[0][1].numpy()[0]
    return output

  return run_model
